#!/usr/bin/env python

"""Faster charmap codec decode and encode in Pyrex

Fastcharmap is about 5 or 6 times faster than the standard charmap routines.

Usage:  fastcharmap.hook('codec_name')

Much magic happens, and calls to the codec's decode and encode will be faster. 
Fast translation caches for decode and encode are built from the mappings when
the codec is called.  If the mappings are changed the cache is rebuilt on the
next call.

If there is a problem, the original codec routines are used.  If hook() hasn't
been called, it is a problem.  If the cache can't be built, it is a problem. 
If Python changes the structure of charmap codecs, fastcharmap will not hook,
and the original codec routines will be used.

When calling the codec, "errors" must be one of nothing, "strict", "ignore", or
"replace", or any error becomes a NotImplementedError.  Also, "ignore" is
slower.

On decode, U+FFFD REPLACEMENT CHARACTER is considered an "error" and handled
according to "errors"; when the cache is built any undefined inputs become
U+FFFD.  On encode, having no mapping is considered an "error" and handled
according to "errors".

As fastcharmap is an extension module not part of the standard Python
distribution, it may not always be available.  Here is one way to cope:

    try:
        import fastcharmap
    except ImportError:
        print "fastcharmap not available"
    else:
        fastcharmap.hook('codec_name')

Here's a sample run using the "rot_13" charmap codec:

>>> chatty(2)
>>> unicode('Some plain ASCII text', 'rot_13')  # without fastcharmap
u'Fbzr cynva NFPVV grkg'
>>> hook('rot_13')                              # now with fastcharmap
3
>>> unicode('Some plain ASCII text', 'rot_13')  # makes cache first time
_make_decode_cache
u'Fbzr cynva NFPVV grkg'
>>> unicode('Some plain ASCII text', 'rot_13')  # just uses cache second time
u'Fbzr cynva NFPVV grkg'
>>> _.encode('rot_13')                          # makes cache first time
_make_encode_cache
'Some plain ASCII text'
>>> _.encode('rot_13')          # just uses cache second time (and overrotates)
'Fbzr cynva NFPVV grkg'
>>> _.encode('iso8859_10')      # an unhooked codec just uses the original code
_orig_encode (not hooked)
'Fbzr cynva NFPVV grkg'
"""

__author__  = "Tony Nelson"
__email__   = "tonynelson@georgeanelson.com"
__url__     = "http://georgeanelson.com/fastcharmap.htm"
__copyright__ = "Copyright 2005 by George A. Nelson.  All rights reserved."
__license__ = "LGPL"
__date__    = "16 Oct 2005"
__version__ = "0.1.1"

# This file is part of fastcharmap.
# 
# Fastcharmap is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# Fastcharmap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with fastcharmap; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import types, sys, codecs

try:
    _orig_decode = codecs.charmap_decode
    _orig_encode = codecs.charmap_encode
except:
    _orig_decode = None
    _orig_encode = None
_chatty = [0]

def chatty(n):
    """Set how chatty.
    
    0: quiet (default)
    1: print a message whenever a cache is rebuilt
    2: and when fast coding couldn't be done
    3: and on every mapping dict access
    """
    _chatty[0] = n

class _ProxyMethodWrapper:
    """Method wrapper object to clear obj._cache.
    """
    def __init__( self, obj, func ):
        self.obj, self.func = obj, func
        assert obj is not None
        assert func is not None

    def __call__( self, *args, **kwds ):
        self.obj._cache = None
        return self.func(*args, **kwds)

class _DictHook(object):
    """Proxy object for dicts to hook dict methods that change the contents.
    """
    def __init__(self, obj):
        self._obj = obj
        self._cache = None
        assert isinstance(self._obj, dict)

    def __getattribute__(self, name):
        """Return a proxy wrapper object if this is a method call.
        """
        if _chatty[0] >= 3:
            print "__getattribute__(%s)" % name
        forward = True
        if name.startswith('_'):
            forward = False
            if name.startswith('__'):
                forward = True
                if name in [ '__dict__', '__getattribute__', '__init__',
                             '__new__', '__weakref__' ]:
                    forward = False
        if not forward:
            return object.__getattribute__(self, name)
        
        att = getattr(self._obj, name)
        if name in [ '__delitem__', '__setitem__',
                     'clear', 'pop', 'popitem', 'setdefault', 'update' ]:
            att = _ProxyMethodWrapper(self, att)
        return att

    # These are needed to forward [] syntax.
    #
    def __getitem__(self, key):
        return self.__getattribute__('__getitem__')(key)

    def __setitem__(self, key, value):
        return self.__getattribute__('__setitem__')(key, value)

    def __delitem__(self, key):
        return self.__getattribute__('__delitem__')(key)

def _make_decode_cache(decoding_map):
    """The decode cache is a direct map from char to unicode, as a unicode
    string.  Returns a cache or False if the cache can't be built, to prevent
    multiple rebuild attempts.
    """
    if _chatty[0] >= 1:
        print "_make_decode_cache"
    cache = [u'\uFFFD'] * 256     # fill with U+FFFD REPLACEMENT CHARACTER
    for ci, ui in decoding_map.items():
        if ui is None:
            continue
        if ( not isinstance(ci, int) or not isinstance(ui, int)
          or not (0 <= ci <= 0xFF) or not (0 <= ui <= 0xFFFF) ):
            return False
        cache[ci] = unichr(ui)
    return u''.join(cache)

def _make_encode_cache(encoding_map):
    """The encode cache is a hash table with external chaining from unicode to
    char, as a unicode string.  Returns a cache or False if the cache can't be
    built, to prevent multiple rebuild attempts.
    """
    # The hash function is the low 8 bits of unicode.  The cache is a unicode
    # string with 256 entry hash table and 256 chain list locations as pairs of
    # (unicode, ((next << 8) + char)), as if:
    #
    #   cdef union Hash:
    #       Py_UNICODE              unused
    #       unsigned char           next_chain
    #   
    #   cdef struct Chain:
    #       Py_UNICODE              input_unichar
    #       union:
    #           Py_UNICODE          unused
    #           struct:
    #               unsigned char   next_chain
    #               char            output_char
    #   
    #   cdef struct HashTable:
    #       Hash                    hash[256]
    #       Chain                   chain[256]
    #
    # Most chains are expected to have one element.  Py_UNICODE is 16 or more
    # bits, and we need 8 bits for an output char, leaving 8 bits for a next
    # link, so we only allow 256 unicode characters.  This is normally the max
    # anyway, as charmap codecs make the mapping by inverting the decoding_map.
    #
    # Umm, the values of next_chain are a bit odd.  In the hash table, there is
    # need for a null value and up to 256 "next" values, from 1..256.  In the
    # chains, the next_chain is stored in 8 bits, with no room for value 256. 
    # This is OK because chain elements are allocated from 0 up, so the highest
    # value that can appear in a chain's next_chain is 255, for element 254.
    
    if _chatty[0] >= 1:
        print "_make_encode_cache"
    cache = [u'\x00'] * (3 * 0x100)
    chaini = 0
    for ui, ci in encoding_map.items():
        if ui is None:
            continue
        if ( not isinstance(ci, int) or not isinstance(ui, int)
          or not (0 <= ci <= 0xFF) or not (0 <= ui <= sys.maxunicode) ):
            return False
        if chaini >= 0x100:
            return False
        hashi = ui & 0xFF
        li = chaini*2 + 0x100
        next = ord(cache[hashi])
        cache[hashi] = unichr(chaini+1)
        cache[li] = unichr(ui)
        cache[li+1] = unichr((next << 8) | ci)
        chaini += 1
    return u''.join(cache)

def hook(codec_name):
    """Install fast hooks into a charmap codec.  Try to do nothing on failure.
    Returns 3 for success:  1*(decode installed) + 2*(encode installed), or 0 if
    couldn't install anything or already installed.  Return False for failure.
    """
    if not _orig_decode or not _orig_encode:
        return False
    n = 0
    encoder = codecs.getencoder(codec_name)
    try:
        codec = encoder.im_self
        if not isinstance(codec, codecs.Codec):
            return False
    
        module = __import__( codec.__module__, globals(), locals(),
            ['decoding_map', 'encoding_map'] )
        import _fastcharmap
    except:
        return False
    
    # Eww.  Instead of having _fastcharmap import fastcharmap by name, we shove
    # the definitions into _fastcharmap.  This keeps the two linked even when we
    # have a different name, such as __main__ (or when importing via a different
    # path).
    if not getattr(_fastcharmap, '_DictHook', None):
        for name in [ '_DictHook', '_make_decode_cache', '_make_encode_cache',
                      '_orig_decode', '_orig_encode', '_chatty' ]:
            setattr(_fastcharmap, name, globals()[name])
    
    decoding_map = getattr(module, "decoding_map", None)
    if ( decoding_map and not isinstance(decoding_map, _DictHook)
     and isinstance(decoding_map, dict) ):
        module.decoding_map = _DictHook(decoding_map)
        codecs.charmap_decode = _fastcharmap._decode
        n =+ 1
    encoding_map = getattr(module, "encoding_map", None)
    if ( encoding_map and not isinstance(encoding_map, _DictHook)
     and isinstance(encoding_map, dict) ):
        module.encoding_map = _DictHook(encoding_map)
        codecs.charmap_encode = _fastcharmap._encode
        n += 2
    
    return n

def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()
